
;====================
; DXBEACON project
; Controller of HAM radio beacon
; (C) Pavel Vachal,OK1DX 2011
; version 1.0 AT Mega 8
; 
; HAMWARE - free for Amateur radio community
; commercial use only with author's approval
;====================
; history: initial release 1.0

.NOLIST

;*********  FUNCTION LIST  *********
; LOOP	- main program loop, started after initialisation

;******** INTERRUPT HANDLERS *******
; RESET - entry point
; TIM0_OVF - timer 0 overflow: handling of PLL control voltage signal
; ADCQRV - AD converter data available

;******** FUSES ********************
; ATmega8 needs to program fuses to operates correctly with external Xtal resonator
; SUT1 = 1, SUT0 = 1, CKSEL3 = 1, CKSEL2 = 1, CKSEL1 = 1, CKSEL0 = 1
; higher fuses byte = 11111111
; fuses programming must be done separately (not performed during code programming)!!!!

.include	"..\..\include\m8def.inc"
.LIST

; port initialisation
; DDR 1=output, 0=input
; PORT when input 1=pullup 0=tri state
; Xtal freq = 12 MHz
.EQU	_DDRB	=0x7		; PB0-PB2 out, PB3-PB5 inp pullup, PB6-7 inp
.EQU	_DDRBX	=0x4		; PB0,1 in, PB2 out, PB6-7 inp
.EQU	_PORTB	=0x38
.EQU	_DDRC	=0x0	; PC0-5 inputs, pullup
.EQU	_PORTC	=0x3F
.EQU	_DDRD	=0x04	; PD0-2 inputs pullup, PD3 inverted key output, PD4 counter input PD5-7 inputs pullup
.EQU	_PORTD	=0xEF

; static RAM registers
.EQU	CWPTR = 0x60	; EEPROM CW text pointer
.EQU	CWCHAR = 0x61	; character being sent
.EQU	FRQ_LO = 0x62	; freq constant, low byte
.EQU	FRQ_HI = 0x63	; freq constant, high byte
.EQU	SHIFT = 0x64	; freq 0=max shift, 15=min shift 

; registers
.DEF	CWSPD   =R1	; speed constant from EEPROM
.DEF	CWCLK	=R2	; speed clock counter (up to CWSPD)
.DEF	T1LO	=R3	; last value read from TIMER1 register
.DEF	T1HI	=R4
.DEF	ISREG	=R15	; temp SREG storage during processing of INT handler
.DEF	TMP1	=R16	; general temp regs used in main loop
.DEF	TMP2	=R17
.DEF	TMP3	=R18
.DEF	ITMP1	=R19	; general temp regs used in INT handlers
.DEF	ITMP2	=R20
.DEF	ITMP3	=R21
.DEF	ITMP4	=R22
.DEF	CWREG	=R23	; bit 7 set when T2 expired, bit 6 elem being sent, 0-3 bits counters
.DEF	FREQREG	=R24	; bit 7 freq corr direction 1=UP 0=DWN; bit 6 FSK key; 0-3 bits FSK counter
.DEF	FREQDEV =R25	; freq deviation, absol. value (0 = no deviation)

;********* MORSE CODE TABLE ********
.EQU	_A	=	0b110
.EQU	_B	=	0b10001
.EQU	_C	=	0b10101
.EQU	_D	=	0b1001
.EQU	_E	=	0b10
.EQU	_F	=	0b10100
.EQU	_G	=	0b1011
.EQU	_H	=	0b10000
.EQU	_I	=	0b100
.EQU	_J	=	0b11110
.EQU	_K	=	0b1101
.EQU	_L	=	0b10010
.EQU	_M	=	0b111
.EQU	_N	=	0b101
.EQU	_O	=	0b1111
.EQU	_P	=	0b10110
.EQU	_Q	=	0b11011
.EQU	_R	=	0b1010
.EQU	_S	=	0b1000
.EQU	_T	=	0b11
.EQU	_U	=	0b1100
.EQU	_V	=	0b11000
.EQU	_W	=	0b1110
.EQU	_X	=	0b11001
.EQU	_Y	=	0b11101
.EQU	_Z	=	0b10011
.EQU	_0	=	0b111111
.EQU	_1	=	0b111110
.EQU	_2	=	0b111100
.EQU	_3	=	0b111000
.EQU	_4	=	0b110000
.EQU	_5	=	0b100000
.EQU	_6	=	0b100001
.EQU	_7	=	0b100011
.EQU	_8	=	0b100111
.EQU	_9	=	0b101111
.EQU	_SL	=	0b101001	; /
.EQU	_AR	=	0b101010
.EQU	_SK	=	0b1101000
.EQU	_PT	=	0b1101010	; .
.EQU	_CM	=	0b1110011	; ,
.EQU	_EQ	=	0b110001	; =

.EQU	_SPC	=	1		; space
.EQU	_TEND	=	0	; end of text

.CSEG
.ORG	0

; interrupt vector table	
	rjmp	RESET	;1	Reset Handler
	reti 		;2 IRQ0 Handler
	reti		;3 IRQ1 Handler
	rjmp	TIM2_COM;4 TIMER2_COMP
	rjmp	TIM2_OVF;5 TIMER2_OVF
	reti		;6 rjmp TIM1_CAPT ; Timer1 Capture Handler
	reti		;7 rjmp TIM1_COMPA ; Timer1 compare Handler
	reti		;8 rjmp TIM1_COMPA ; Timer1 compare Handler
	reti		;9 Timer1 Overflow Handler
	rjmp	TIM0_OVF;10 Timer0 Overflow Handler
	reti		;11 SPI Transfer Complete Handler
	rjmp 	UART_RXC ;12 UART RX Complete Handler
	rjmp 	UART_DRE ;13 UDR Empty Handler
	reti		;14 rjmp UART_TXC ; UART TX Complete Handler
	rjmp	ADCQRV	;15 ADC ; ADC Conversion Complete Interrupt Handler
	reti		;16 EEPROM Ready Handler
	reti		;17 Analog Comparator Handler
	reti		;18 TWI
	reti		;19 SPM_RDY
	
RESET:
; initiate stack pointer
	ldi	TMP1,0x5F		; end of RAM of Mega8
	out	SPL,TMP1		; stack
	ldi	TMP1,4
	out	SPH,TMP1

	ldi	TMP1,_DDRB	; port mode ini
	out	DDRB,TMP1
	ldi	TMP1,_DDRC
	out	DDRC,TMP1
	ldi	TMP1,_DDRD
	out	DDRD,TMP1
	ldi	TMP1,_PORTB
	out	PORTB,TMP1
	ldi	TMP1,_PORTC
	out	PORTC,TMP1
	ldi	TMP1,_PORTD
	out	PORTD,TMP1
	
; register INI
	clr	CWREG
	ldi	TMP1,CWTXT
	sts	CWPTR,TMP1
	ldi	TMP1,1
	sts	CWCHAR,TMP1
	clr	TMP1
	clr	T1LO
	clr	T1HI
	clr	FREQREG
	
	
; params from EEPROM
	clr	TMP2
	ldi	TMP1,ECWSPD
	call	RD_EE
	mov	CWSPD, TMP1	; CW speed constant
	mov	CWCLK, TMP1
	
	ldi	TMP1, ESHIFT
	call	RD_EE
	sts	SHIFT,TMP1	; shift value
	
	ldi	TMP1, EFREQ
	call	RD_EE
	sts	FRQ_LO,TMP1	; freq constant low byte
	
	ldi	TMP1, EFREQ+1
	call	RD_EE
	sts	FRQ_HI,TMP1	; freq constant high byte
		
	
; TIMER 0, 1, 2 initialisation
;	ldi	TMP1,7
	ldi	TMP1,3			; !!! SIMULATION ONLY internal clock 1:64 !!!
	out	TCCR0,TMP1		; T0 8bit - using external clock, no prescaler, rising edge
	ldi	TMP1,1
	out	TCCR1B,TMP1		; T1 16bit - internal osc, no prescaler, rising edge
	clr	TMP1
	out	TCCR1A,TMP1
	ldi	TMP1,0x4
	out	TCCR2,TMP1		; T2 8bit - internal/64
	ldi	TMP1,0xFF
	out	OCR2,TMP1
	
	ldi	TMP1,0xC1		; TOIE0 = 1, TOIE2 = 1, OCIE2 = 1 - different bits in Mega8!!!
	out	TIMSK,TMP1		; enable both TIMER0 and TIMER1 overflow interrupt
	
	sei					;enable global interrupt; the fun start NOW!

;********** MAIN PROGRAM LOOP *********
;it waits for events signalised by interrupt rutines
	
LOOP:

; CW message transmission
	sbrs	CWREG,7			; CW time expired
	rjmp	LOOP2
	cbr	CWREG,0x80		; clear the flag
	mov	TMP1,CWREG
	andi	TMP1, 0x0f		; get counter
	breq	CW1			; counter empty
	dec	CWREG
	rjmp 	LOOP2			; not yet empty, further wait
	
CW1:	sbrs	CWREG,6			; ? element over, insert 1 space
	rjmp	CW2
	ldi	CWREG,0x00		; KEY DWN after dot or dash elem = 0, length = 1
	cbi	PORTB,2			; KEY = 0
	sbi	PORTD,3
	cbr	FREQREG,6
	rjmp	LOOP2
	
CW2:	lds	TMP1,CWCHAR
	cpi	TMP1,1
	breq	CW3			; ? send space after char
	cpi	TMP1,0
	breq	CW4			; ? take next char
	
	lsr	TMP1
	brcs	CW5
	ldi	CWREG,0x40		; send DOT , length = 1
	rjmp	CW6
CW5:	ldi	CWREG,0x42		; send DASH , length = 3
CW6:	sbi	PORTB,2			; KEY = 1
	cbi	PORTD,3
	sbr	FREQREG,6
	sts	CWCHAR,TMP1
	rjmp	LOOP2
	
CW3:	ldi	CWREG,0x40		; SPACE AFTER CHAR elem = true, length = 1
	clr	TMP1
	sts	CWCHAR,TMP1
	rjmp	LOOP2
	
CW4:	clr	TMP2
	lds	TMP1,CWPTR
	mov	TMP3,TMP1
	call	RD_EE			; read one char from EEPROM
	inc	TMP3
	sts	CWPTR,TMP3
	sts	CWCHAR,TMP1
	cpi	TMP1,_TEND
	brne	CW2			; ? end of text
	ldi	TMP1,CWTXT		; set pointer to begin
	sts	CWPTR,TMP1
	rjmp	CW4			; read again from begining

		
LOOP2:	nop

	rjmp	LOOP	;back in main loop

;======================================== end of main loop

		
;****** INTERRUPT HANDLER FOR UART RX COMPLETE *******
; spare
UART_RXC:
		in	ISREG,SREG

RX2:	out	SREG,ISREG
		reti
		
;****** INTERRUPT HANDLER FOR UART TX EMPTY ********
; spare
UART_DRE:
		in ISREG,SREG

SND2:		out SREG,ISREG
		reti
		
		
;****** INTERRUPT HANDLER OF TIMER 0 (8 bit) **********
; every 256 external clock pulses (VCO freq divided by 512)
; for 50.1 MHz VCO interrupt freq 382 Hz (2.62 msec)
; calculate difference of TIMER1 counter since last TIM0_OVF
; subtract the difference with freq constant
; if the result is zero, no freq correction needed (freq exact)	
; if the result is POSITIVE, it means we need to tune UP (the T0 period too long)
; if the result is NEGATIVE we need to tune DOWN
; the result is further evaluated - if >255 or < -255 we set max correction possible (0xFF]
; if between -255 and 255 we set the correction value to abs(value)
; FSK realized by increment of "deviation value" when key is down
; this increment is not performed every cycle, just once every SHIFT cycle (counter FREQREG bits 0-3)
TIM0_OVF:
; **** calculate TIMER1 pulses between subseqent TIM0_OVF
		in	ISREG,SREG
		in	ITMP1,TCNT1L
		in	ITMP2,TCNT1H		; new counter 1 value
		mov	ITMP3,ITMP1
		mov	ITMP4,ITMP2		; save temporarily
		sub	ITMP1,T1LO
		sbc	ITMP2,T1HI		; current value - last cycle value
		mov	T1LO,ITMP3
		mov	T1HI,ITMP4		; save new counter value as old
; **** subtract the freq constant	
		lds	ITMP3,FRQ_LO
		sub	ITMP1, ITMP3
		lds	ITMP3,FRQ_HI
		sbc	ITMP2,ITMP3
; **** DO FSK KEYING ****		
		sbrs	FREQREG,6		; ? key is down
		rjmp	TIM0_6
		mov	ITMP3,FREQREG
		andi	ITMP3,0x0F		; ? bits 0-3 == 0
		breq	TIM0_5
		dec	FREQREG			; decrement 0-3 bits, no other action
		rjmp	TIM0_4			; no other action
TIM0_5:		lds	ITMP3,SHIFT
		add	FREQREG,ITMP3		; set bits 0-3 to shift
		ldi	ITMP3,1
		add	ITMP1,ITMP3
		clr	ITMP3
		adc	ITMP2,ITMP3		; result += 1 (do shift)
		rjmp	TIM0_4
TIM0_6:		andi	FREQREG,0xF0		; key up - just clear bits 0-3 for future
; ITMP1 = deviation LSB, ITMP2 = deviation MSB with sign
; calculate freq correction direction and deviation 8bit value
TIM0_4:		sbrs	ITMP2,7			; ? highest bit of MSB
		rjmp	TIM0_1
		cbr	FREQREG,7		; set flag correction DWN
		cpi	ITMP2,0xFF		; the upper byte = 0xFF, just small correction
		brne	TIM0_2
		and	ITMP1,ITMP1		; 0 -> max correction
		breq	TIM0_2
		neg	ITMP1			; invert sign (=absolute value)
		rjmp	TIM0_3
TIM0_1:		sbr	FREQREG,7		; set flag correction UP
		and	ITMP2,ITMP2		; ? high byte = 0
		breq	TIM0_3
TIM0_2:		ldi	ITMP1,0xFF		; do max possible correction
TIM0_3:		mov	FREQDEV, ITMP1		; save new value of freq deviation
TIM0_END:	out	SREG,ISREG
		reti


; ****** TIMER 2 INTERRUPT HANDLER ****** performed every 5.46 miliseconds
; used for CW keying - other handling in main loop
; used to set PLL outputs (set pin as output to level 0 or 1 as needed)
; there are 2 PLL outputs, one is inverted the other
TIM2_OVF:
		in	ISREG,SREG
		dec	CWCLK
		brne	TIM2_1
		mov	CWCLK,CWSPD
		sbr	CWREG,0x80	; set flag - time expired
		
TIM2_1:		and	FREQDEV,FREQDEV	; no deviation, no action
		out	OCR2,FREQDEV
		breq	TIM2_2
		ldi	ITMP1,_DDRB	; set ports PB0 and PB1 as outputs
		out	DDRB,ITMP1
		sbrc	FREQREG,7	; freq correction direction
		rjmp	TIM2_1
		sbi	PORTB,0		; set PB0	/ tune UP
		cbi	PORTB,1		; clear PB1
		rjmp	TIM2_2
		cbi	PORTB,0		; set PB0	/ tune DWN
		sbi	PORTB,1		; clear PB1
		
TIM2_2:		out	SREG,ISREG
		reti
		
; ****** TIMER 2 COMPARE INTERRUPT HANDLER ******
; set both PLL outputs to high impedance state (end of RC integration)
TIM2_COM:
		in	ISREG,SREG
		ldi	ITMP1,_DDRBX
		out	DDRB,ITMP1		; set ports PB0 and PB1 as Hi impedance inputs (no pullup)
		cbi	PORTB,0
		cbi	PORTB,1
		out	SREG,ISREG
		reti

;****** INTERRUPT HANDLER OF ANALOG / DIGITAL CONVERTER *******
; for future implementation
ADCQRV:					; called when AD conversion finished
		reti
			
;******** PROCEDURE reads one byte from EEPROM
RD_EE:	; input: LSB address in TMP1, MSB in TMP2; output: byte in TMP1
	out	EEARH,TMP2
	out	EEARL,TMP1
	sbi	EECR,EERE
RDEE1:	sbic	EECR,EERE
	rjmp	RDEE1
	in	TMP1,EEDR	; text position byte
	ret
		
.ESEG
.ORG	0x0

VERS:	.DB		'D','x','B','e','a','c','o','n',' ','1','.','0',' ','O','K','1','D','X'
;EFREQ:	.DB		0xC2, 0x7A	; frequency constant = 131072 * XTAL / QRG ; 50.05 MHz = 31426 = 0x7AC2
EFREQ:	.DB		0x00, 0x40	; simulation 0x4000
ECWSPD:	.DB		40, 0
ESHIFT:	.DB		4, 0		; 0=max shift, 15=min shift
CWTXT:	.DB		_O, _K, _1, _D, _X, _SL, _B, _SPC, _J, _N, _6, _9, _S, _D, _SPC, _TEND